package pa;

import java.nio.IntBuffer;

import org.lwjgl.BufferUtils;
import org.lwjgl.PointerBuffer;
import org.lwjgl.opencl.CL10;
import org.lwjgl.opencl.CLCommandQueue;
import org.lwjgl.opencl.CLContext;
import org.lwjgl.opencl.CLKernel;
import org.lwjgl.opencl.CLMem;
import org.lwjgl.opencl.CLProgram;

import pa.cl.CLUtil;
import pa.cl.CLUtil.PlatformDevicePair;
import pa.cl.OpenCL;
import pa.util.BufferHelper;
import pa.util.IOUtil;
import pa.util.SizeOf;
import pa.util.math.MathUtil;

public class Scan
{
    
    private static CLContext context;
    private static CLKernel scan;
    private static CLKernel spreadSums;
    private static CLMem input;
    private static CLCommandQueue queue;
    private static CLProgram program;
    
    private static boolean useShrd = true;
    
    public static void main(String[] args)
    {
        CLUtil.createCL();
        
        PlatformDevicePair pair = CLUtil.choosePlatformAndDevice();
        
        context = OpenCL.clCreateContext(pair.platform, pair.device, null, null);
        queue = OpenCL.clCreateCommandQueue(context, pair.device, CL10.CL_QUEUE_PROFILING_ENABLE);
        program = OpenCL.clCreateProgramWithSource(context, IOUtil.readFileContent("kernel/scan.cl"));
        OpenCL.clBuildProgram(program, pair.device, "-cl-nv-verbose", null);
        scan = OpenCL.clCreateKernel(program, useShrd ? "scan" : "scanGM");
        spreadSums = OpenCL.clCreateKernel(program, "spreadSums");
        
        int localWorkSize = 512;
        
        int size = 1 << 27;
        System.out.println("Elements="+size);
        IntBuffer data = BufferUtils.createIntBuffer(size);
        for(int i = 0; i < size; ++i)
        {
            data.put(MathUtil.nextInt(2));
        }
        //data.put(7, 1);

        //BufferHelper.printBuffer(data, size);
        
        data.position(0);
        
        input = OpenCL.clCreateBuffer(context, CL10.CL_MEM_READ_ONLY | CL10.CL_MEM_COPY_HOST_PTR, data);
        
        PointerBuffer workSize = new PointerBuffer(1);
        workSize.put(0, size / 2);
        
        PointerBuffer lws = new PointerBuffer(1);
        lws.put(0, localWorkSize);
        
        int sumsCnt = (int)(workSize.get(0) / lws.get(0));
        
        long scanTime = 0, scanSumsTime = 0, addSumsTime = 0;
        
        CLMem sums = OpenCL.clCreateBuffer(context, 0, sumsCnt * SizeOf.INTEGER);
        
        OpenCL.clSetKernelArg(scan, 0, input);
        OpenCL.clSetKernelArg(scan, 1, sums);
        OpenCL.clSetKernelArg(scan, 2, 1);
        
        if(useShrd)
        {
          OpenCL.clSetKernelArg(scan, 3, (long)(2*localWorkSize * SizeOf.INTEGER));   
        }
        
        scanTime = CLUtil.measureKernelCall(scan, queue, workSize, lws, 1);
        //OpenCL.clEnqueueReadBuffer(queue, input, 1, 0, data, null, null);
        
        //BufferHelper.printBuffer(data, size);
      
        OpenCL.clSetKernelArg(scan, 0, sums);
        OpenCL.clSetKernelArg(scan, 2, 0);
        
        if(useShrd)
        {
          OpenCL.clSetKernelArg(scan, 3, (long)sumsCnt * SizeOf.INTEGER);   
        }
        
        if(sumsCnt <= 2048)
        {
            workSize.put(0, sumsCnt / 2);
            
            scanSumsTime = CLUtil.measureKernelCall(scan, queue, workSize, workSize, 1);

            OpenCL.clSetKernelArg(spreadSums, 0, sums);
            OpenCL.clSetKernelArg(spreadSums, 1, input);
            OpenCL.clSetKernelArg(spreadSums, 2, 2*localWorkSize);
            workSize.put(0, size - localWorkSize);
            lws.put(0, localWorkSize);
            //System.out.println(workSize.get(0) + "," + localWorkSize);
            addSumsTime = CLUtil.measureKernelCall(spreadSums, queue, workSize, null, 1);
        }
 
        /*IntBuffer sumsTmp = BufferUtils.createIntBuffer(sumsCnt);
        
        OpenCL.clEnqueueReadBuffer(queue, sums, 1, 0, sumsTmp, null, null); 
        
        BufferHelper.printBuffer(sumsTmp, sumsCnt);
        */
        //OpenCL.clEnqueueReadBuffer(queue, input, 1, 0, data, null, null);
        
        //BufferHelper.printBuffer(data, size);
        System.out.println(String.format("SubScan=%f\nScanSums=%f\nAddSums=%f\nScan=%f", scanTime/1e9, scanSumsTime/1e9, addSumsTime/1e9, (scanTime + scanSumsTime + addSumsTime) /1e9));
        
        OpenCL.clReleaseMemObject(sums);
        OpenCL.clReleaseMemObject(input);
        OpenCL.clReleaseCommandQueue(queue);
        OpenCL.clReleaseKernel(scan);
        OpenCL.clReleaseProgram(program);
        OpenCL.clReleaseContext(context);
        
        CLUtil.destroyCL();
    }
}
