kernel void scanGM(global int* data, local int* shrdMem, global int* sums, const int writeSums)
{
    int thid = get_local_id(0);
    int grpId = get_group_id(0);
    int id = get_global_id(0);
    int grpSize = 2 * get_local_size(0);
    int N = grpSize;
    int offset = 1;
       
    for(int d = N>>1; d > 0; d >>= 1)
    {   
        barrier(CLK_GLOBAL_MEM_FENCE);
        if (thid < d)  
        {
            int ai = grpId * grpSize + offset*(2*thid+1)-1;  
            int bi = grpId * grpSize + offset*(2*thid+2)-1;
            data[bi] += data[ai];
        }
        offset *= 2;
    }
    
    if (thid == get_local_size(0) - 1) { data[2*id+1] = 0; }

    for(int d = 1; d < N; d *= 2)
    {  
        offset >>= 1;  
        barrier(CLK_GLOBAL_MEM_FENCE);
        if (thid < d)                       
        {  
            int ai = grpId * grpSize + offset*(2*thid+1)-1;  
            int bi = grpId * grpSize + offset*(2*thid+2)-1;
            int t = data[ai];  
            data[ai] = data[bi];  
            data[bi] += t;   
        } 
    }
    
    barrier(CLK_GLOBAL_MEM_FENCE);
    
    if(writeSums && (thid == get_local_size(0)-1))
    {
        sums[grpId] = data[2*id+1];
    } 
}


kernel void scan(global int* data, local int* shrdMem, global int* sums, const int writeSums)
{
    int thid = get_local_id(0);
    int grpId = get_group_id(0);
    int grpSize = 2 * get_local_size(0);
    int N = grpSize;
    int offset = 1;
    
    shrdMem[2*thid  ] = data[2*thid + grpId * grpSize];
    shrdMem[2*thid+1] = data[2*thid + 1 + grpId * grpSize]; 
    int last = shrdMem[2*thid+1];
    for(int d = N>>1; d > 0; d >>= 1)
    {   
        barrier(CLK_LOCAL_MEM_FENCE);
        if (thid < d)  
        {
            int ai = offset*(2*thid+1)-1;  
            int bi = offset*(2*thid+2)-1;
            shrdMem[bi] += shrdMem[ai];
        }
        offset *= 2;
    }
    
    if (thid == 0) { shrdMem[N - 1] = 0; }
        
    for(int d = 1; d < N; d *= 2)
    {  
        offset >>= 1;  
        barrier(CLK_LOCAL_MEM_FENCE);
        if (thid < d)                       
        {  
            int ai = offset*(2*thid+1)-1;  
            int bi = offset*(2*thid+2)-1;  
            int t = shrdMem[ai];  
            shrdMem[ai] = shrdMem[bi];  
            shrdMem[bi] += t;   
        } 
    }
    
    barrier(CLK_LOCAL_MEM_FENCE);
    data[2*thid + grpId * grpSize] = shrdMem[2*thid];
    data[2*thid + 1 + grpId * grpSize] = shrdMem[2*thid+1];
    
    if(writeSums && (thid == get_local_size(0)-1))
    {
        sums[grpId] = shrdMem[2*thid+1] + last;
    }
}

kernel void spreadSums(global int* sums, global int* data)
{
    int grpSize = get_local_size(0);
    int grpId = get_group_id(0);
    int gid = get_global_id(0);
    
    int sum = sums[grpId+1];
    data[grpSize + gid] += sum;
}

